#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import re
import inspect

import pymongo

from . import datatypes
from ..settings import database


def _to_snake(name: str) -> str:
    name = re.sub("([A-Z])", r"_\1", name)
    return name.lower().strip("_")


class Model(abc.ABC):
    HOST = None
    PORT = None
    DATABASE = None
    COLLECTION = None

    def __init__(self, **key_value):
        cls = self.__class__

        # if "__data" not in dir(cls):
        #     cls.__data = []
        #     cls.__primary_key = None

        #     for name in dir(self):
        #         ty = getattr(self, name)
        #         if isinstance(ty, datatypes.ModelTypeBase):
        #             cls.__data.append((name, ty))
        #             if ty.primary:
        #                 cls.__primary_key = (name, ty)
        cls._setup()

        for name, ty in cls.__data:
            value = ty.default
            if name in key_value:
                value = ty.create_instance(key_value[name])
            setattr(self, name, value)

    @classmethod
    def _setup(cls):
        if "__data" not in dir(cls):
            cls.__data = []
            cls.__primary_key = None
            col = cls.__collection()

            indices = col.list_indexes()

            for name in dir(cls):
                ty = getattr(cls, name)
                if isinstance(ty, datatypes.ModelTypeBase):
                    cls.__data.append((name, ty))
                    if ty.primary:
                        cls.__primary_key = (name, ty)
                    if ty.unique:
                        if name in indices:
                            col.create_index(name, unique=True)

    @classmethod
    def __collection(cls):
        if "__collection_name" not in dir(cls):
            if cls.COLLECTION is None:
                cls.__collection_name = _to_snake(cls.__name__)
            else:
                cls.__collection_name = cls.COLLECTION
        if cls.HOST is None:
            cls.HOST = database.host
        if cls.PORT is None:
            cls.PORT = database.port

        if cls.DATABASE is None:
            cls.DATABASE = database.database

        client = pymongo.MongoClient(host=cls.HOST, port=cls.PORT, tz_aware=True)
        db = client[cls.DATABASE]
        return db[cls.__collection_name]

    @classmethod
    def load(cls, multi=False, primary=None, **key_value):
        cls._setup()
        query = key_value
        if primary is not None:
            name, _ = cls.__primary_key
            query = {name: primary}

        if not multi:
            data = cls.__collection().find_one(query)
            if data is None:
                return None
            return cls(**data)

        return [cls(**data) for data in cls.__collection().find(query)]

    def save(self) -> bool:
        cls = self.__class__
        primary_name, _ = cls.__primary_key
        data = {}
        for name, ty in cls.__data:
            datum = ty.to_model_data(getattr(self, name))
            if datum is None and ty.non_null:
                return False
            data[name] = datum
        res = self.__collection().update_one({primary_name: data[primary_name]}, {"$set": data}, upsert=True)
        return res.acknowledged

    @classmethod
    def get_primary_key_name(cls):
        cls._setup()
        name, _ = cls.__primary_key
        return name


__imported_by = []


def load_loaded_models():
    classes = []
    for imported_by in __imported_by:
        classes.extend(inspect.getmembers(imported_by, inspect.isclass))  # lambda obj: issubclass(type(obj), Model)))
    return [cls for _, cls in classes if cls != Model if issubclass(cls, Model)]


__loaded_models = None


def get_loaded_models():
    global __loaded_models
    if __loaded_models is None:
        models = load_loaded_models()
        __loaded_models = {cls.__name__: cls for cls in models}
        __loaded_models.update({_to_snake(name): cls for name, cls in __loaded_models.items()})
    return __loaded_models


def get_loaded_model(name):
    return get_loaded_models()[name]


class TestData(Model):
    name = datatypes.StringType(default="cerussite")
    age = datatypes.IntType()


if __name__ == '__main__':
    test = TestData()
    print(test.name)
    get_loaded_models()
else:
    __imported_by.extend([inspect.getmodule(stack[0]) for stack in inspect.stack()])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import hashlib
import base64
import re
import typing


__V1_STRETCH_COUNT = 10000


def __version0(target: str) -> str:
    """
    hash function version 0
    :param target: input data
    :return: hashed data
    """
    bytes_ = target.encode(encoding="utf-8")
    salt = [
        base64.b16encode(bytes_).decode(encoding="utf-8"),
        base64.b32encode(bytes_).decode(encoding="utf-8"),
        base64.b64encode(bytes_).decode(encoding="utf-8"),
        base64.b85encode(bytes_).decode(encoding="utf-8")
    ]

    def to_salt(d, s):
        return (d + s + d + s + d + s).encode(encoding="utf-8")

    for i in range(__V1_STRETCH_COUNT):
        s512 = hashlib.sha512()
        s512.update(to_salt(target, salt[i % 4]))
        target = s512.hexdigest()

    return "$0$" + target


__HASH_FUNCTIONS = [
    __version0
]


def compute_hash(target: str, ver=None) -> str:
    """
    compute hash value
    :param target: input data
    :param ver: hash function version. None is latest version
    :return: hashed data
    """
    if ver is None:
        ver = -1
    return __HASH_FUNCTIONS[ver](target)


def latest_version() -> int:
    """
    latest hash function version
    :return: latest version number
    """
    return len(__HASH_FUNCTIONS) - 1


def check(hashed: str, non_hashed: str) -> typing.Tuple[bool, typing.Optional[str]]:
    """
    check correct value and get updated hashed value
    :param hashed: value 1
    :param non_hashed: value 2
    :return: (is corrected, updated value)
    """
    ver = version(hashed)
    non_ver = version(non_hashed)
    new_value = None

    if ver is None:
        if non_ver is None:
            hashed = compute_hash(hashed)
            non_hashed = compute_hash(non_hashed)
            new_value = hashed
        elif non_ver != latest_version():
            new_value = compute_hash(hashed)
        hashed = compute_hash(hashed, non_ver)
    else:
        if non_ver is None:
            if ver != latest_version():
                new_value = compute_hash(non_hashed)
            non_hashed = compute_hash(non_hashed, ver)
        else:
            if non_ver == ver:
                return hashed == non_hashed, None
            raise RuntimeError("check function cannot do check hashed value and hashed value")

    if hashed != non_hashed:
        return False, None

    return True, new_value


def version(target: str) -> typing.Optional[int]:
    """
    get hashed data version
    :param target: hashed data
    :return: hashed data version
    """
    if target is None:
        return None
    match = re.match("\$(\d+)\$.+", target)
    if not match:
        return None
    return int(match.group(1))


class Hashed:
    """
    Hashed data type
    """
    def __init__(self, data=None, ver=None):
        """
        constructor
        :param data: data
        :param ver: version
        """
        if data is None:
            data = ""
        elif version(data) is None:
            data = compute_hash(data, ver)
        self.__data = data

    def __str__(self):
        """
        str operator
        hashed data
        :return: hashed data
        """
        return self.__data

    def __eq__(self, other):
        """
        equal operator
        :param other: opponent value
        :return: is same
        """
        if isinstance(other, Hashed):
            other = str(other)
        ok, new = check(self.__data, other)

        if new is not None:
            self.__data = new
        return ok



from pymvc import model, settings, router, wsgi, controller

settings.database.database = "flask_mvt_test"
app = wsgi.app


class User(model.Model):
    name = model.StringType()
    id = model.UniqueIdType(primary=True, non_null=True, unique=True)


class TestData1(model.Model):
    pass


class TestData2(model.Model):
    pass


class TestData3(model.Model):
    pass


class TestData4(model.Model):
    pass


user = User.load(primary="6f15de0e-deb0-4368-bdf1-21a87734fc14")
user.name = "new name"
user.save()


# @app.route('/')
# def hello_world():
#     return render_template("test.html")

@router.route("/")
class TopController(controller.Controller):
    VIEW = "test.html"

    def get(self, **kwargs):
        pass


# router.add_route("/", TopController)


if __name__ == '__main__':
    app.run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
from flask import abort


class RenderParameter:
    """
    parameter for render_template
    """
    def __init__(self, view: str, **params):
        """
        constructor
        :param view: view file name
        :param params: parameters
        """
        self.__view = view
        self.__params = params

    @property
    def view(self):
        """
        get view file name
        :return: view file name
        """
        return self.__view

    @property
    def params(self):
        """
        get render parameters
        :return: render parameters
        """
        return self.__params


def render(view=None, **params):
    """
    get parameter class
    :param view: view file name (optional)
    :param params: template parameters (optional)
    :return: parameter instance
    """
    return RenderParameter(view, **params)


class Controller(abc.ABC):
    """
    Controller base class
    """
    VIEW = None

    def get(self, **kwargs):
        """
        http GET function
        :param kwargs: url parameter
        :return: response
        """
        return abort(404)

    def post(self, **kwargs):
        """
        http POST function
        :param kwargs: url parameter
        :return: response
        """
        return abort(404)

    def put(self, **kwargs):
        """
        http PUT function
        :param kwargs: url parameter
        :return: response
        """
        return abort(404)

    def delete(self, **kwargs):
        """
        http DELETE function
        :param kwargs: url parameter
        :return: response
        """
        return abort(404)

    def __hash__(self):
        """
        hash function
        :return: hash value of Controller name
        """
        return hash(self.__class__.__name__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .model_base import Model
from .datatypes import StringType, IntType, FloatType, UniqueIdType
from .datatypes import ForeignType, EnumType, ListType, BoolType, DatetimeType, HashType

from . import binder



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import typing
import abc
import uuid
import datetime
import enum


class ModelTypeBase(abc.ABC):
    """
    Model data type.
    """

    def __init__(self, primary: bool=False, non_null: bool=False, unique: bool=False, default=None):
        """
        Initializer
        :param primary: is primary key
        :param non_null: not null value
        :param unique: unique value
        """
        self.__primary = primary
        self.__non_null = non_null
        self.__unique = unique
        if non_null and default is None:
            raise TypeError("type specified non null but default value is None or not specified")

        if primary:
            self.__unique = True
            self.__non_null = True

        if default is not None and not isinstance(default, self.type):
            raise TypeError(
                "default value type is not compatible with specified type.\n"
                "type(default)={}, specified type={}".format(type(default), self.type)
            )
        self.__default = default

    @property
    def primary(self) -> bool:
        """
        is primary key
        :return: is primary key
        """
        return self.__primary

    @property
    def non_null(self) -> bool:
        """
        is non null
        :return: is non null
        """
        return self.__non_null

    @property
    def unique(self) -> bool:
        """
        is unique value
        :return: is unique value
        """
        return self.__unique

    @property
    @abc.abstractmethod
    def type(self) -> typing.Type:
        """
        value type
        :return: value type
        """
        pass

    def create_instance(self, value):
        """
        create specified instance
        :param value: value
        :return: data instance
        """
        return value

    def to_model_data(self, value):
        """
        convert to model data
        :param value: original value
        :return: data
        """
        return value

    @property
    def default(self):
        """
        get default value
        :return: default value
        """
        return self.__default


class StringType(ModelTypeBase):
    """
    String type
    """
    @property
    def type(self) -> typing.Type:
        return str


class IntType(ModelTypeBase):
    """
    integer type
    """
    @property
    def type(self) -> typing.Type:
        return int


class FloatType(ModelTypeBase):
    """
    float type
    """
    @property
    def type(self) -> typing.Type:
        return float


class UniqueIdType(ModelTypeBase):
    """
    UUID data type
    """
    def __init__(self, primary: bool=False, non_null: bool=False, unique: bool=False, default: uuid.UUID=uuid.uuid4()):
        super(UniqueIdType, self).__init__(primary, non_null, unique, default)

    @property
    def type(self):
        return uuid.UUID

    def create_instance(self, value):
        return uuid.UUID(value)

    def to_model_data(self, value):
        return str(value)


class ListType(ModelTypeBase):
    """
    List type
    """
    def __init__(self, dt: typing.Type, *args,
                 primary: bool=False, non_null: bool=False, unique: bool=False, default=None, **kwargs):
        super(ListType, self).__init__(primary, non_null, unique, default)

        from .model_base import Model
        if issubclass(dt, Model):
            self.__data_instance = ForeignType(dt)
        elif issubclass(dt, enum.Enum):
            self.__data_instance = EnumType(dt)
        elif issubclass(dt, ModelTypeBase):
            self.__data_instance = dt(*args, **kwargs)
        self.__data_type = dt

    @property
    def type(self) -> typing.Type:
        return list

    def create_instance(self, value):
        if issubclass(self.__data_type, ModelTypeBase):
            return [self.__data_instance.create_instance(v) for v in value]
        return value

    def to_model_data(self, value):
        if issubclass(self.__data_type, ModelTypeBase):
            return [self.__data_instance.to_model_data(v) for v in value]
        return value


class ForeignType(ModelTypeBase):
    """
    Foreign model value type
    """
    def __init__(self, model, model_primary_key=None,
                 primary: bool=False, non_null: bool=False, unique: bool=False, default=None):
        """
        constructor
        :param model: model name or model class
        :param model_primary_key: primary key (optional)
        :param primary: this item is primary
        :param non_null: non null value
        :param unique: unique value
        :param default: default value
        """
        super(ForeignType, self).__init__(primary, non_null, unique, default)

        from pymvc.model import model_base as mb

        if isinstance(model, str):
            model = mb.get_loaded_model(model)
        self.__model = model

        if model_primary_key is None:
            model_primary_key = self.__model.get_primary_key_name()
        self.__model_primary_key = model_primary_key

    def create_instance(self, value):
        return self.__model.load(**{self.__model_primary_key: value})

    def to_model_data(self, value):
        return getattr(value, self.__model_primary_key)

    @property
    def type(self):
        return self.__model


class EnumType(ModelTypeBase):
    """
    enum.Enum model value type
    """
    def __init__(self, enum_type, primary: bool=False, non_null: bool=False, unique: bool=False, default=None):
        super(EnumType, self).__init__(primary, non_null, unique, default)

        self.__enum = enum_type

    def create_instance(self, value):
        return self.__enum(value)

    def to_model_data(self, value):
        return value.value

    @property
    def type(self):
        return self.__enum


class DatetimeType(ModelTypeBase):
    """
    datetime.datetime model value type
    """
    @property
    def type(self):
        return datetime.datetime


class BoolType(ModelTypeBase):
    """
    bool model value type
    """
    @property
    def type(self):
        return bool


class HashType(ModelTypeBase):
    """
    hashed model value type
    """
    @property
    def type(self):
        from .hash_function import Hashed
        return Hashed

    def create_instance(self, value):
        return self.type(value)

    def to_model_data(self, value):
        return str(value)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from . import wsgi
import typing
import pathlib


class _DatabaseSetting:
    """
    database setting class
    """
    def __init__(self):
        self.__host = None
        self.__port = None
        self.__database = None

    @property
    def host(self) -> typing.Optional[str]:
        """
        get host name
        :return: host name
        """
        return self.__host

    @host.setter
    def host(self, value: typing.Optional[str]):
        """
        set host name
        :param value: host name
        """
        self.__host = value

    @property
    def port(self) -> typing.Optional[int]:
        """
        get port number
        :return: port number
        """
        return self.__port

    @port.setter
    def port(self, value: typing.Optional[int]):
        """
        set port number
        :param value: port number
        """
        self.__port = value

    @property
    def database(self) -> typing.Optional[str]:
        """
        get database name
        :return: database name
        """
        return self.__database

    @database.setter
    def database(self, value: str):
        """
        set database name
        :param value: database name
        """
        self.__database = value


class _FlaskSetting:
    """
    Flask settings
    """
    def __init__(self):
        cwd = str(pathlib.Path.cwd().absolute())
        self.root_directory = cwd

    @property
    def static_directory(self) -> str:
        """
        get static files directory
        :return: static files directory
        """
        return wsgi.app.static_folder

    @static_directory.setter
    def static_directory(self, value: str):
        """
        set static files directory
        :param value: static files directory
        """
        wsgi.app.static_folder = value

    @property
    def static_url(self) -> str:
        """
        get static files url
        :return: static files url
        """
        return wsgi.app.static_url_path

    @static_url.setter
    def static_url(self, value: str):
        """
        set static files url
        :param value: static files url
        """
        wsgi.app.static_url_path = value

    @property
    def templates_directory(self) -> str:
        """
        get templates directory
        :return: templates directory
        """
        return wsgi.app.template_folder

    @templates_directory.setter
    def templates_directory(self, value: str):
        """
        set templates directory
        :param value: templates directory
        """
        wsgi.app.template_folder = value

    @property
    def root_directory(self) -> str:
        """
        get Flask root path
        :return: root path
        """
        return wsgi.app.root_path

    @root_directory.setter
    def root_directory(self, value: str):
        """
        set Flask root path
        :param value: root path
        """
        wsgi.app.root_path = value


database = _DatabaseSetting()
flask = _FlaskSetting()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .. import wsgi
from .model_base import get_loaded_model


@wsgi.app.context_processor
def model_loader():
    """
    Flask (Jinja2) function for loading model data
    :return: Flask (Jinja2) functions
    """
    def load_one(model, primary=None, **query):
        """
        load one model's data
        :param model: model class name or snake case class name
        :param primary: primary key (optional)
        :param query: query (optional)
        :return: data or None
        """
        model = get_loaded_model(model)
        model = model.load(primary=primary, **query)
        return model

    def load_many(model, primary=None, **query):
        """
        load model's data
        :param model: model class name or snake case class name
        :param primary: primary key (optional)
        :param query: query (optional)
        :return: data list
        """
        model = get_loaded_model(model)
        model = model.load(multi=True, primary=primary, **query)
        return model

    return {"load_one": load_one, "load_many": load_many}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .pkg_info import __copyright__, __version__, __license__, __author__

from flask import request, jsonify, abort, redirect, session

from .controller import Controller, render
from .router import route, add_route, url_for
from .model import Model, binder
from .model import UniqueIdType, IntType, FloatType, StringType, ForeignType, EnumType, ListType
from .model import DatetimeType, BoolType, HashType
from .wsgi import app
from . import settings



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

from pymvc import pkg_info

with open("README.md") as fp:
    readme = fp.read()

with open("requirements.txt") as fp:
    requirements = [r.strip() for r in fp.readlines()]


package_name = "pymvc"

setup(
    name=package_name,
    packages=find_packages(),
    version=pkg_info.__version__,
    license=pkg_info.__license__,
    author=pkg_info.__author__,
    url="https://github.com/SiLeader/pymvc",

    description="MVC framework for Python",
    long_description=readme,
    long_description_content_type="text/markdown",

    install_requires=requirements,

    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware"
    ]
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import typing

from flask import render_template
from flask import url_for as flask_url_for

from . import wsgi, controller


__routes = {}


def add_route(path: str, ctrl: typing.Union[controller.Controller, typing.Type]):
    """
    add router class to system
    :param path: url (see Flask reference)
    :param ctrl: control class or class instance
    """
    if not isinstance(ctrl, controller.Controller):
        ctrl = ctrl()

    def vf(func):
        def internal(**kwargs):
            res = func(**kwargs)

            if res is None or isinstance(res, controller.RenderParameter):
                view = ctrl.VIEW
                params = {}
                if res is not None:
                    params = res.params
                    if res.view is not None:
                        view = res.view

                res = render_template(view, **params)
            return res
        return internal

    endpoint = "{}:{}".format(path, ctrl.__class__.__name__)
    __routes[ctrl.__class__.__name__] = endpoint+":GET"

    wsgi.app.add_url_rule(path, view_func=vf(ctrl.get), methods=["GET"], endpoint=endpoint+":GET")
    wsgi.app.add_url_rule(path, view_func=vf(ctrl.post), methods=["POST"], endpoint=endpoint+":POST")
    wsgi.app.add_url_rule(path, view_func=vf(ctrl.put), methods=["PUT"], endpoint=endpoint+":PUT")
    wsgi.app.add_url_rule(path, view_func=vf(ctrl.delete), methods=["DELETE"], endpoint=endpoint+":DELETE")


def route(path: str):
    """
    add router class to system as decorator
    :param path: url (see Flask reference)
    """
    def route_deco(cls):
        add_route(path, cls)
    return route_deco


def url_for(ctrl: typing.Union[controller.Controller, str, typing.Type], **kwargs):
    """
    get url for Controller
    :param ctrl: Controller instance or name
    :param kwargs: url fragment data
    :return: url for Controller
    """
    if not isinstance(ctrl, str):
        if isinstance(ctrl, controller.Controller):
            ctrl = ctrl.__class__.__name__
        else:
            ctrl = ctrl.__name__
    return flask_url_for(__routes[ctrl], **kwargs)


@wsgi.app.context_processor
def url_for_function():
    """
    add url_for to Jinja2
    :return: Jinja2 function
    """
    return {"url_for": url_for}
